/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "server.h"
#include "util/memory.h"
#include "assert/assert.h"
#include "console/console.h"
#include "util/string.h"

server_t SERVER;

void cmdStart(const consolecmdexec_t *exec) {
  serverstart_t start;

  if(exec->argc > 0) {
    if(stringCompare(exec->argv[1], "0") == 0) {
      start.type = SERVER_TYPE_SINGLE_PLAYER;
    } else if(stringCompare(exec->argv[0], "1") == 0) {
      start.type = SERVER_TYPE_NETWORKED;
    } else {
      consolePrint("Invalid server type: %s", exec->argv[0]);
      return;
    }
  } else {
    start.type = SERVER_TYPE_NETWORKED;
  }

  if(exec->argc > 1) {
    if(start.type == SERVER_TYPE_NETWORKED) {
      if(!stringToU16(exec->argv[1], &start.networked.port)) {
        consolePrint("Invalid port number: %s", exec->argv[0]);
        return;
      }
    }
  } else {
    if(start.type == SERVER_TYPE_NETWORKED) {
      start.networked.port = SERVER_DEFAULT_PORT;
    }
  }

  errorret_t ret = serverStart(start);
  if(ret != ERROR_OK) {
    consolePrint("Failed to start server: %s", errorString());
    errorFlush();
    return;
  }
}

void cmdClose(const consolecmdexec_t *exec) {
  serverStop();
}

void serverInit() {
  memoryZero(&SERVER, sizeof(server_t));
  SERVER.state = SERVER_STATE_STOPPED;

  consoleRegCmd("start", cmdStart);
  consoleRegCmd("close", cmdClose);
}

errorret_t serverStart(const serverstart_t start) {
  errorret_t ret;
  assertIsMainThread("Server start must be on main thread");

  // Do not start a running server.
  if(SERVER.state != SERVER_STATE_STOPPED) {
    return error("Server is already running");
  }

  SERVER.type = start.type;

  // Hand off to relevant server type to start.
  switch(start.type) {
    case SERVER_TYPE_NETWORKED:
      ret = networkedServerStart(&SERVER, start);
      break;

    default:
      assertUnreachable("Invalid server type");
  }

  if(ret != ERROR_OK) SERVER.state = SERVER_STATE_STOPPED;
  return ret;
}

uint8_t serverGetClientCount() {
  uint8_t i = 0, clientCount = 0;
  do {
    if(SERVER.clients[i].state == SERVER_CLIENT_STATE_DISCONNECTED) continue;
    clientCount++;
  } while(i < SERVER_MAX_CLIENTS);
  return clientCount;
}

void serverStop() {
  assertIsMainThread("Server stop must be on main thread");
  if(SERVER.state == SERVER_STATE_STOPPED) return;

  consolePrint("Stopping server...");
  switch(SERVER.type) {
    case SERVER_TYPE_NETWORKED:
      networkedServerStop(&SERVER);
      break;

    default:
      assertUnreachable("Invalid server type");
  }
  consolePrint("Server stopped.");
}

void serverDispose() {
  serverStop();
}