/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "serverclient.h"
#include "server/networked/networkedserver.h"

#define SERVER_MAX_CLIENTS 32
#define SERVER_DEFAULT_PORT 3030

typedef enum {
  SERVER_STATE_STOPPED,
  SERVER_STATE_STARTING,
  SERVER_STATE_RUNNING,
  SERVER_STATE_STOPPING
} serverstate_t;

typedef enum {
  SERVER_TYPE_SINGLE_PLAYER,
  SERVER_TYPE_NETWORKED
} servertype_t;

typedef struct serverstart_s {
  servertype_t type;
  union {
    networkedserverstart_t networked;
  };
} serverstart_t;

typedef struct server_s {
  serverstate_t state;
  servertype_t type;
  pthread_t thread;
  serverclient_t clients[SERVER_MAX_CLIENTS];
  
  union {
    networkedserver_t networked;
  };
} server_t;

extern server_t SERVER;

/**
 * Initialize the server
 * 
 * This function initializes the server by setting up the ENet library and
 * creating a server host.
 */
void serverInit();

/**
 * Start the server
 * 
 * This function starts the server by creating a host, binding it to the
 * specified address and port, and starting the server thread.
 * 
 * @param port The port number to bind the server to.
 * @return Returns an error code if the server fails to start.
 */
errorret_t serverStart(const serverstart_t start);

/**
 * Get the client count
 * 
 * This function returns the number of connected clients to the server.
 * 
 * @return The number of connected clients.
 */
uint8_t serverGetClientCount();

/**
 * Stop the server
 * 
 * This function stops the server by destroying the host and cleaning up
 * resources.
 */
void serverStop();

/**
 * Dispose of the server
 * 
 * This function disposes of the server by deinitializing the ENet library.
 */
void serverDispose();